Imports System.Math
Imports System.text

Module MultiPrecision10
    Public Const RXdc As Integer = 4
    Public Const RX As Short = CShort(10 ^ RXdc)

    Private Const RXm As Short = RX - 1

    Public Sub MpADD(ByVal u() As Short, ByVal pu As Integer, _
                     ByVal v() As Short, ByVal pv As Integer, ByVal m As Integer, _
                     ByVal w() As Short)
        pu += m - 1
        pv += m - 1

        Dim a(m) As Short
        Dim pa As Integer = m

        Dim ireg As Short = 0
        Do
            ireg = ireg \ RX + u(pu) + v(pv)
            a(pa) = ireg Mod RX
            pu -= 1
            pv -= 1
            pa -= 1
        Loop Until pa = 0
        a(0) = ireg \ RX

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpSUB(ByVal u() As Short, ByVal pu As Integer, _
                     ByVal v() As Short, ByVal pv As Integer, ByVal m As Integer, _
                     ByVal w() As Short, ByRef isign As Integer)
        pu += m - 1
        pv += m - 1

        Dim a(m - 1) As Short
        Dim pa As Integer = m - 1

        Dim ireg As Short = RX
        Do
            ireg = ireg \ RX + RXm + u(pu) - v(pv)
            a(pa) = ireg Mod RX
            pu -= 1
            pv -= 1
            pa -= 1
        Loop Until pa < 0
        isign = ireg \ RX - 1

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpNEG(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer)
        Dim ou As Integer = pu
        pu += m - 1

        Dim ireg As Short = RX
        Do
            ireg = RXm - u(pu) + ireg \ RX
            u(pu) = ireg Mod RX
            pu -= 1
        Loop Until pu < ou
    End Sub

    Public Sub MpSAD(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal iv As Integer, ByVal w() As Short)
        pu += m - 1

        Dim a(m) As Short
        Dim pa As Integer = m

        Dim ireg As Short = u(pu) + CShort(iv)
        a(pa) = ireg Mod RX
        pu -= 1
        pa -= 1

        Do Until pa = 0
            ireg = u(pu) + ireg \ RX
            a(pa) = ireg Mod RX
            pu -= 1
            pa -= 1
        Loop

        a(0) = ireg \ RX

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpSMU(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal iv As Integer, ByVal w() As Short)
        pu += m - 1

        Dim a(m) As Short
        Dim pa As Integer = m

        Dim ireg As Integer = 0
        Do
            ireg = u(pu) * iv + ireg \ RX
            a(pa) = CShort(ireg Mod RX)
            pu -= 1
            pa -= 1
        Loop Until pa = 0
        a(0) = CShort(ireg \ RX)

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpSDV(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal iv As Integer, ByVal q() As Short, ByRef ir As Integer)
        Dim a(m - 1) As Short
        Dim pa As Integer = 0

        ir = 0
        Do
            Dim i As Integer = RX * ir + u(pu)
            a(pa) = CShort(i \ iv)
            ir = i Mod iv

            pu += 1
            pa += 1
        Loop Until pa = m

        a.CopyTo(q, 0)
        Erase a
    End Sub

    Public Sub MpMUL(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal v() As Short, ByVal pv As Integer, ByVal n As Integer, _
                     ByVal w() As Short)
        Dim n_max As Integer = Max(m, n)
        Dim nm As Integer = n + m - 1

        Dim nn As Integer = 1
        Do While (nn < n_max)
            nn <<= 1
        Loop
        nn <<= 1


        Dim a(nn - 1) As Double
        For J As Integer = 0 To m - 1
            a(J) = u(pu + J)
        Next

        Dim b(nn - 1) As Double
        For J As Integer = 0 To n - 1
            b(j) = v(pv + j)
        Next

        rdft(a, FFT_FORWARD)
        rdft(b, FFT_FORWARD)

        b(0) *= a(0)
        b(1) *= a(1)
        For re As Integer = 2 To nn - 1 Step 2
            Dim bre As Double = b(re)
            Dim im As Integer = re + 1
            b(re) = bre * a(re) - b(im) * a(im)
            b(im) = bre * a(im) + b(im) * a(re)
        Next

        rdft(b, FFT_REVERSE)

        Dim cy As Long = 0
        For j As Integer = nn - 1 To 0 Step -1
            Dim t As Long = CLng(b(j) / (nn >> 1)) + cy
            cy = t \ RX
            If j < nm Then w(j + 1) = CShort(t Mod RX)
        Next

        If cy > RXm Then Throw New System.ArithmeticException("cannot happen in fftmul")
        w(0) = CShort(cy)

        Erase a, b
    End Sub

    Public Sub MpINV(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, ByVal v() As Short)
        Const MF As Integer = 16 \ RXdc
        Const BI As Double = 1 / RX
        Const RXm2 As Short = RX - 2

        Dim n As Integer = v.Length

        '================================================================
        'Use ordinary floating arithmetic to get an initial approximation
        '================================================================
        Dim mm As Integer = Min(MF, m)
        Dim fu As Double = u(pu + mm - 1)
        For j As Integer = mm - 2 To 0 Step -1
            fu *= BI
            fu += u(pu + j)
        Next

        Dim fv As Double = 1 / fu

        Dim v1(2 * n - 1) As Short
        Dim rpv1 As Integer = 0

        For j As Integer = 0 To n - 1
            Dim i As Double = Floor(fv)
            v1(j) = CShort(i)
            fv = RX * (fv - i)
        Next

        '=====================================
        'Iterate Newton's rule to convergence.
        '=====================================
        Dim s(n + m - 1) As Short
        Do
            '---------------------
            'Construct 2 - VU in S
            '---------------------
            's = v1 * u
            's : n + m digits with the radix point after s(1)
            MpMUL(v1, rpv1, n, u, pu, m, s)

            'calculate 2 - s with n digits
            'the radix point is located after s(1)
            MpNEG(s, 1, n)
            s(1) -= RXm2

            '------------------
            'Multiply SV into V
            '------------------
            'v1 = s * v1
            'v1 : n + n digits with the radix point after v1(1)
            MpMUL(s, 1, n, v1, rpv1, n, v1)
            rpv1 = 1

            '--------------------------------------------------------------
            'If fractional part of S is not zero, it has not converged to 1
            '--------------------------------------------------------------
            Dim j As Integer
            For j = 2 To n - 1
                If s(j) > 0 Then Exit For
            Next

            If j = n Then Exit Do
        Loop

        '==========================================
        'store the most significant n digits into V
        '==========================================
        Array.Copy(v1, 1, v, 0, n)
        Erase s, v1
    End Sub

    Public Sub MpSQRT(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                      ByVal v() As Short, ByVal w() As Short)
        Const MF As Integer = 16 \ RXdc
        Const BI As Double = 1 / RX

        Dim n As Integer = w.Length
        Dim mm As Integer = Min(MF, m)

        Dim fu As Double = u(pu + mm - 1)
        For j As Integer = mm - 2 To 0 Step -1
            fu *= BI
            fu += u(pu + j)
        Next

        Dim fw As Double = 1 / Sqrt(fu)

        Dim w1(2 * n - 1) As Short
        For j As Integer = 0 To n - 1
            Dim i As Double = Floor(fw)
            w1(j) = CShort(i)
            fw = RX * (fw - i)
        Next

        MySQRT(u, pu, m, v, w, w1)

        Erase w1
    End Sub

    Public Sub MpSQRT(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                      ByVal v() As Short, ByVal w() As Short, ByVal wi() As Short)
        Dim n As Integer = w.Length
        Dim w1(2 * n - 1) As Short

        For j As Integer = 0 To n - 1
            w1(j) = wi(j)
        Next

        MySQRT(u, pu, m, v, w, w1)

        Erase w1
    End Sub

    Private Sub MySQRT(ByVal u() As Short, ByVal pu As Integer, ByVal m As Integer, _
                       ByVal v() As Short, ByVal w() As Short, ByVal w1() As Short)
        Const RXm3 As Short = RX - 3

        Dim n As Integer = w.Length
        Dim rpw1 As Integer = 0

        '=====================================
        'Iterate Newton's rule to convergence.
        '=====================================
        Dim s(n + Max(n, m) - 1) As Short
        Do
            '-----------------------------
            'Construct S = (3 - U W W) / 2
            '-----------------------------
            's = w * w
            's : n + n digits with the radix point after r(1)
            MpMUL(w1, rpw1, n, w1, rpw1, n, s)

            's = u * s
            '  = u * w * w
            's : n + m digits with the radix point after s(1)
            '    only n digits of s are used.
            MpMUL(s, 1, n, u, pu, Min(m, n), s)

            'calculate 3 - s with n digits
            'the radix point is located after s(1)
            MpNEG(s, 1, n)
            s(1) -= RXm3

            'divide by 2
            'the radix point is located after s(0)
            Dim ir As Integer
            MpSDV(s, 1, n, 2, s, ir)

            '--------------------------------------------------------------
            'If fractional part of S is not zero, it has not converged to 1
            '--------------------------------------------------------------
            Dim j As Integer
            For j = 1 To n - 2
                If s(j) > 0 Then Exit For
            Next

            If j = n - 1 Then Exit Do

            '--------------------
            'Replace w1 by s * w1
            '--------------------
            'w1 = s * w1
            'n digits of s and n digits of w1 produces
            'n + n digits of w1.
            'the radix point is located after t(1)
            MpMUL(s, 0, n, w1, rpw1, n, w1)
            rpw1 = 1
        Loop

        '==========================================
        'Get square root from reciprocal and return
        '==========================================
        'v = s = u * w1
        'v = 1/sqrt(u) * u = sqrt(u)
        MpMUL(w1, rpw1, n, u, pu, Min(m, n), s)

        Array.Copy(w1, rpw1, w, 0, n)
        Array.Copy(s, 1, v, 0, n)
    End Sub

    Public Sub MpDIV(ByVal u() As Short, ByVal v() As Short, _
                     ByRef q() As Short, ByRef r() As Short)
        Const MACC As Integer = 6

        Dim n As Integer = u.Length
        Dim m As Integer = v.Length

        Dim u1(n - 1) As Short
        u.CopyTo(u1, 0)
        Dim v1(m - 1) As Short
        v.CopyTo(v1, 0)

        '-------------
        'Set S = 1 / V
        '-------------
        Dim s(n + MACC - 1) As Short
        MpINV(v, 0, m, s)

        '-----------
        'Set Q = S U
        '-----------
        'rr = s * u : n + n + MACC bytes
        '  s : n + MACC bytes
        '  u : n bytes
        Dim rr(2 * n + MACC - 1) As Short
        MpMUL(s, 0, n + MACC, u, 0, n, rr)
        MpSAD(rr, 0, n + MACC - 1, 1, s)

        If (q Is Nothing) OrElse (q.Length < n - m + 1) Then ReDim q(n - m)
        Array.Copy(s, 2, q, 0, n - m + 1)

        '------------------------------------------
        'Multiply and subtract to get the remainder
        '------------------------------------------
        MpMUL(q, 0, n - m + 1, v1, 0, m, rr)
        Dim isign As Integer
        MpSUB(u1, 0, rr, 1, n, rr, isign)
        If (r Is Nothing) OrElse (r.Length < m) Then ReDim r(m - 1)
        Array.Copy(rr, n - m, r, 0, m)

        Erase s, rr, u1, v1
    End Sub
End Module
